import os
import arrow
import hashlib
import time
import inquirer
import click
import sys
import subprocess
from pathlib import Path
from collections import namedtuple
from importlib import import_module
from contextlib import contextmanager
import importlib.util

PROC = namedtuple("Process", field_names=("process", "path", "md5"))

def _get_md5(filepath):
    if not filepath.exists():
        return ''
    m = hashlib.md5()
    m.update(filepath.read_bytes())
    return m.hexdigest()

def _raise_error(msg):
    click.secho(msg, fg='red')
    sys.exit(-1)

def _get_bots_paths(config):
    bots_paths = []

    for path in config.config.get('bots-paths', []):
        if not path: continue
        path = Path(path)
        for path in path.glob("**/*.py"):
            if '.git' in path.parts:
                continue
            if path.name.startswith("__"): continue
            if path.name.endswith(".py"):
                bots_paths.append(path.parent)

    for path in bots_paths:
        if path not in sys.path:
            sys.path += [path]
    return bots_paths

def iterate_scripts(config):
    result = set()

    def _collect():
        for bots_path in _get_bots_paths(config):
            for script in bots_path.glob("*.py"):
                if script.name.startswith("__"):
                    continue
                yield script
    for x in _collect():
        result.add(x)
    return sorted(list(result))

def _get_robot_file(config, name):
    if name and name.startswith("/") or name.startswith("./"):
        name = Path(name).absolute()
    if not name:
        scripts = list(sorted(set(iterate_scripts(config))))
        questions = [
            inquirer.List('file', choices=scripts)
        ]
        answer = inquirer.prompt(questions)
        if not answer['file']:
            _raise_error("No file selected")
        return answer['file']

    filtered = list(filter(lambda x: name in x.name, iterate_scripts(config)))
    if len(filtered) > 1:
        _raise_error(f"Too many bots found for {name}")
        return
    if not filtered:
        _raise_error(f"No bot found for {name}")
        return
    return filtered[0].absolute()

def kill_proc(proc, timeout):
    from . import global_data
    if not global_data['config']:
        return
    p_sec = 0
    for second in range(timeout):
        if proc.process.poll() is None:
            time.sleep(0.1)
            p_sec += 1
    if p_sec >= timeout:
        proc.process.kill() # supported from python 2.6
    config = global_data['config']
    config.processes.pop(config.processes.index(proc))

def kill_all_processes():
    from . import global_data
    timeout_sec = 1
    if not global_data['config']:
        return
    for p in global_data['config'].processes:
        kill_proc(p, timeout_sec)

def cleanup():
    kill_all_processes()

def _select_bot_path():
    from . import global_data
    config = global_data['config']
    if not config.config.get('bots-paths'):
        _raise_error(f"Please configure a path in {config.config_file}")
    if len(config.config.get('bots-paths')) == 1:
        return config.config.get('bots-paths')[0]
    questions = [
        inquirer.List(name='path', message="Path", choices=sorted(config.config['bots-paths'])),
    ]
    answer = inquirer.prompt(questions)
    if not answer:
        return
    path = answer['path']
    return path

def load_module(path):
    mod_name = path.name.rsplit(".", 1)[0]
    spec = importlib.util.spec_from_file_location(mod_name, path)
    foo = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(foo)
    return foo

@contextmanager
def _onetime_client(name_appendix, script, timeout=10):
    from .mqtt_tools import _get_regular_client
    from .mqtt_tools import _connect_client
    data = {'stop': False}
    def on_publish(client, userdata, mid):
        data['stop'] = True

    client = _get_regular_client(name=name_appendix, scriptfile=script)
    client.on_publish = on_publish
    _connect_client(client)
    yield client
    timeout = arrow.get().shift(seconds=timeout)
    while not data['stop'] and arrow.get() < timeout:
        client.loop(0.1)
    client.disconnect()


import traceback
import os
from . import global_data
from pathlib import Path
import arrow
import json
import click
import paho.mqtt.client as mqtt

class PseudoClient(object):
    def __init__(self):
        pass

    def publish(self, path, payload=None, qos=0):
        print(f"{path}:{qos}: {payload}")

class mqttwrapper(object):
    def __init__(self, client, hostname, modulename):
        self.client = client
        self.hostname = hostname
        self.logger = global_data['config'].logger
        self.modulename = modulename

    def publish(self, path, payload=None, qos=0, retain=False):
        path = self.hostname + '/' + path
        value = {
            'module': self.modulename,
            'value': payload,
            # cannot use msg.timestamp - they use time.monotonic() which results
            # in consecutive calls results starting from 1970
            'timestamp': str(arrow.get().to('utc')),
        }
        self.output_to_console(path, value)
        self.client.publish(
            path,
            payload=json.dumps(value),
            qos=qos,
            retain=retain
        )

    def output_to_console(self, path, value):
        if os.getenv("SRE_OUTPUT_MESSAGES") != "1":
            return
        try:
            value = json.dumps(value, indent=4)
        except: pass

        click.secho("------------------------------------------------------------------------------------------------------", fg='blue')
        msg = f"{path}:\n{value}"
        click.secho(msg, fg='cyan')

def _get_mqtt_wrapper(client, module):
    from . import global_data
    name = global_data['config'].config['name']
    _name = getattr(module, "HOSTNAME", name) if module else name
    _modulename = Path(module.__file__).stem
    return mqttwrapper(client, _name, _modulename)

def _get_regular_client(name, scriptfile):
    client = mqtt.Client(client_id=f"srebot-{name}-{scriptfile.name})", protocol=mqtt.MQTTv5)
    return client

def _connect_client(client):
    config = global_data['config']
    client.connect(config.config['broker']['ip'], config.config['broker'].get('port', 1883), 60)

def on_connect(client, userdata, flags, reason, properties):
    global_data['config'].logger.debug(f"Client connected")
    client.subscribe("#")

def on_message(client, userdata, msg):
    config = global_data['config']
    config.logger.debug(f"on_message: {msg.topic} {str(msg.payload)}")
    from .tools import load_module
    if not config.bot:
        return
    module = load_module(config.bot)
    if getattr(module, 'on_message', None):
        try:
            client2 = _get_mqtt_wrapper(client, module)
            try:
                value = json.loads(msg.payload)
            except Exception:
                # default values
                value = {
                    'value': msg.payload,
                    'timestamp': str(arrow.get().to('utc')),
                    'module': None,
                }
            module.on_message(client2, msg, value)

        except Exception as ex:
            trace = traceback.format_exc()
            config.logger.error(str(ex) + "\n\n" + str(msg) + '\n\n' + trace)




# SCHEDULERS = ["* * * * * */1"]
# HOSTNAME = 'host1'

def run(client):
    # client.publish('house/bulb1', 'on')
    pass

def on_message(client, msg, payload=None):
    print(msg.topic)


def install():
    pass


#!/usr/bin/env python3
import traceback
import arrow
import stat
import uuid
import signal
import time
import argparse
import inquirer
import sys
from pathlib import Path
import os
import croniter
import subprocess
import json
from datetime import datetime
import threading
import click
from . import cli
from .config import pass_config
from .mqtt_tools import PseudoClient
from .tools import _get_md5, _raise_error, PROC, iterate_scripts, _get_robot_file, kill_proc, _select_bot_path, load_module
from . import global_data
from . mqtt_tools import _get_mqtt_wrapper, _connect_client, _get_regular_client, on_connect, on_message
from .webserver import start_webserver
from .tools import _onetime_client


VERSION = '0.2'

@cli.command(name='new', help="Makes new bot from template")
@click.argument("name", required=True)
@pass_config
def make_new_file(config, name):
    path = _select_bot_path()
    if not path:
        return
    for c in " ":
        name = name.replace(c, "_")
    dest_path = Path(path) / (name + ".py")
    if dest_path.exists():
        _raise_error(f"Already exists: {dest_path}")
    import srebot
    template = (Path(srebot.__file__).parent / 'datafiles' / 'bot.template.py').read_text()
    dest_path.write_text(template)
    click.secho(f"Created new bot in {dest_path}", fg='green')

@cli.command(name="install", help="Installs all new requirements of bots and installs as a system service.")
@click.option("-n", "--name", required=False, default='sre')
@pass_config
def make_install(config, name):
    from . install_services import install_systemd
    from . install_services import install_requirements
    config = global_data['config']

    install_requirements(sys.argv[0])
    if not name.endswith('.service'):
        name = name + ".service"
    install_systemd(name, sys.argv[0])

    paths = ', '.join(config.config['bots-paths'])
    click.secho(f"Add custom bots: {name} new bot1", fg='yellow')
    click.secho(f"{config.config_file}:")
    click.secho(config.config_file.read_text())

@cli.command(name='list')
@pass_config
def list_bots(config):
    for script in iterate_scripts(config):
        click.secho(script, fg='green')

def start_proc(config, path):
    config.logger.info(f"Starting {path}...")
    path = path.absolute()
    os.system(f"pkill -9 -f '{sys.executable} .*{path}'")

    process = subprocess.Popen([
        sys.executable,
        sys.argv[0],
        '-c', config.config_file,
        '-l', config.config.get('log_level'),
        'run',
        path,
    ])
    md5 = _get_md5(path)
    config = global_data['config']
    config.processes.append(PROC(process=process, path=path, md5=md5))

def run_iter(config, client, scheduler, module, once):
    from . import global_data
    global_data['config'] = config
    base = datetime.now()
    iter = croniter.croniter(scheduler, base)
    while True:
        try:
            next = iter.get_next(datetime)
            config.logger.debug(f"Executing {module} next at {next.strftime('%Y-%m-%d %H:%M:%S')}")

            while True:
                if datetime.now() > next or once:

                    client2 = _get_mqtt_wrapper(client, module)
                    script_file = config.bot
                    if getattr(module, 'run', None):
                        try:
                            config.logger.debug(f"Running {script_file}:run")
                            module.run(client2)
                            client2.publish(script_file.name + '/rc', payload=0)
                        except Exception as ex:
                            client2.publish(script_file.name + '/rc', payload=1)
                            client2.publish(script_file.name + '/last_error', payload=str(ex))
                            client2.publish('/last_error', payload=f"{script_file}\n{ex}")
                            config.logger.error(ex)
                            time.sleep(1)
                    break
                time.sleep(0.5)

        except Exception as ex:
            msg = traceback.format_exc()
            config.logger.error(msg)
            time.sleep(1)
        finally:
            if once:
                return

@cli.command(help="Name of script within search directories")
@click.argument("name")
@pass_config
def test_bot(config, name):
    name = _get_robot_file(config, name)

    mod = load_module(name)
    mod.run(PseudoClient())

@cli.command()
@click.argument('path', type=click.Path(exists=False))
@pass_config
def add_bot_path(config, path):
    path = path.strip()
    path = Path(path).absolute()
    path.mkdir(parents=True, exist_ok=True)
    if str(path) not in config.config['bots-paths']:
        config.config['bots-paths'].append(str(path))
    config.store_config()

@cli.command(help="Start main loop or sub daemon script (called by service usually)")
@click.option('-1', '--once', required=False, is_flag=True)
@click.argument('script', required=False)
@pass_config
def run(config, script, once):
    if not script:
        script = _get_robot_file(config, '')
    if not str(script).startswith('/'):
        script = Path(os.getcwd()) / script
        if not script.exists():
            _raise_error("Needs absolute path if killing existing script")

    #if kill_others:
    #    os.system(f"pkill -9 -f '{sys.executable}.*{sys.argv[0]}.*{script}'")
    script = Path(script).absolute()
    module = load_module(script)
    config.logger.info(f"Starting script at {script}")
    config.bot = script
    client = _get_regular_client(config.config['name'], script)
    client.on_connect = on_connect
    client.on_message = on_message

    config.logger.info(f"Connecting to {config.config['broker']}")

    while True:
        try:
            _connect_client(client)
        except Exception as ex:
            msg = traceback.format_exc()
            config.logger.error(msg)
            time.sleep(1)
        else:
            break

    for scheduler in getattr(module, 'SCHEDULERS', []):
        if once:
            with _onetime_client('_run_once', script) as client:
                os.environ["SRE_OUTPUT_MESSAGES"] = "1"
                run_iter(config, client, scheduler, module, once=True)
            return
        else:
            t = threading.Thread(target=run_iter, args=(config, client, scheduler, module, False))
            t.daemon = False
            t.start()
            t.join
    client.loop_forever()

@cli.command(help="Start main loop or sub daemon script (called by service usually)")
@pass_config
def daemon(config):
    config.logger.info("Starting daemon...")
    t = threading.Thread(target=start_webserver,)
    t.daemon = True
    t.start()

    while True:
        for proc in config.processes:
            if _get_md5(proc.path) != proc.md5:
                config.logger.info(f"Script was changed, restarting: {proc.path}")
                kill_proc(proc, 1)
                start_proc(config, proc.path)

        for script in iterate_scripts(config):
            if str(script) in config.config.get('disabled', []):
                continue
            if not [x for x in config.processes if x.path == script]:
                config.logger.info(f"Detected new script: {script}")
                start_proc(config, script)

        for proc in config.processes:
            if not [x for x in list(iterate_scripts(config)) if x == proc.path]:
                config.logger.info(f"Detected removal of script: {script}")
                kill_proc(proc, 1)

        time.sleep(2)

@cli.command(name="state")
@pass_config
def state(config):
    scripts = list(sorted(map(str, iterate_scripts(config))))
    defaults = []
    for script in scripts:
        script = script
        if script not in config.config.get('disabled', []):
            defaults.append(script)
    questions = [inquirer.Checkbox('state', message="Turn on bots", choices=scripts, default=defaults)]
    answer = inquirer.prompt(questions)
    if not answer:
        return

    disabled = []
    for script in scripts:
        if script not in answer['state']:
            disabled.append(script)
    config.config['disabled'] = disabled
    config.store_config()
    click.secho("Settings applied. Dont forget to restart the service.", fg='yellow')


import simplejson
import http.server
import json
from pathlib import Path
import sys

class Handler(http.server.SimpleHTTPRequestHandler) :
    # A new Handler is created for every incommming request tho do_XYZ
    # methods correspond to different HTTP methods.

    def do_GET(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

    def do_POST(self):
        from .tools import _onetime_client
        if self.path != "/":
            self.data_string = self.rfile.read(int(self.headers['Content-Length']))
            data = simplejson.loads(self.data_string)

            if self.path.startswith("/trigger/"):
                with _onetime_client("_webtrigger", Path(sys.argv[0])) as client:
                    data = json.dumps(data).encode('utf-8')
                    client.publish(self.path[1:], data, 2)

        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

def start_webserver():
    from . import global_data
    config = global_data['config'].config
    if not config.get("http_address"):
        global_data['config'].logger.info("No webserver configured - triggering on this host not possible.")
        return

    http_server = http.server.HTTPServer(
        (config['http_address'], int(config.get('http_port', 8822))
    ), Handler)
    http_server.serve_forever()



import os
from . import global_data
from pathlib import Path
import click
import sys
import subprocess
from importlib import import_module
import importlib.util

def _get_exe(config, path_to_sre):
    # get python environment and executable
    options = f" --config-file '{config.config_file}' daemon"
    if os.getenv("VIRTUAL_ENV"):
        ExecStart = (
            f"/bin/sh -c 'cd \\'{os.getenv('VIRTUAL_ENV')}\\' &&"
            f". bin/activate && "
            f"python3 \\'{path_to_sre}\\' {options}'"
        )
    else:
        ExecStart = f"'{sys.executable}' '{path_to_sre}' {options}"
    return ExecStart

def install_systemd(name, path_to_sre):
    config = global_data['config']
    subprocess.call(["systemctl", "stop", name])
    import srebot
    template = (Path(srebot.__file__).parent / 'datafiles' / 'sre.service').read_text()
    ExecStart = _get_exe(config, path_to_sre)

    template = template.replace('__exec_start__', ExecStart)
    template = template.replace('__name__', name)
    path = (Path("/etc/systemd/system/") / name)
    if path.exists():
        click.secho(f"Overwrite existing daemon file: {path}", fg='red')
    path.write_text(template)
    subprocess.check_call(["/bin/systemctl", "daemon-reload"])
    subprocess.check_call(["/bin/systemctl", "enable", name])
    subprocess.check_call(["/bin/systemctl", "restart", name])
    click.secho(f"Successfully installed with systemd: {name}", fg='green')


def install_requirements(path_to_sre):
    # rewrite using virtual env
    from . import global_data
    from .tools import iterate_scripts
    from .tools import load_module
    config = global_data['config']
    for script_path in iterate_scripts(config):
        req_file = script_path.parent / 'requirements.txt'
        if req_file.exists():
            subprocess.check_call([
                sys.executable, "-mpip", "install",
                '-r', req_file])
        mod = load_module(script_path)
        if getattr(mod, 'install', None):
            click.secho(f"Installing {mod.__file__}", fg='green')
            mod.install()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

import io
import json
from glob import glob
import os
import socket
import sys
import subprocess
from shutil import rmtree
from pathlib import Path

import setuptools
from setuptools.config import read_configuration
from setuptools import find_packages, setup, Command
from setuptools.command.install import install
setup_cfg = read_configuration("setup.cfg")
metadata = setup_cfg['metadata']

required_version = (60, 5, 0)
setuptools_version = tuple(map(int, setuptools.__version__.split(".")))
if setuptools_version < required_version:
    raise Exception(f"Requires setup version {'.'.join(map(str, required_version))}")

# HACK to ignore wheel building from pip and just to source distribution
if 'bdist_wheel' in sys.argv:
    sys.exit(0)

NAME = metadata['name']

# Package meta-data.
# What packages are required for this module to be executed?
REQUIRED = [
    "simplejson",
    "paho-mqtt", "click>=8.0.3", "croniter", "arrow", "pudb", "pyyaml", "inquirer",
    "click-completion-helper"
]

# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = metadata['DESCRIPTION']

# Load the package's __version__.py module as a dictionary.
about = {}
if not metadata['version']:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = metadata['version']


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def clear_builds(self):
        for path in ['dist', 'build', NAME.replace("-", "_") + ".egg-info"]:
            try:
                self.status(f'Removing previous builds from {path}')
                rmtree(os.path.join(here, path))
            except OSError:
                pass

    def run(self):
        self.clear_builds()

        self.status('Building Source distribution…')
        os.system('{0} setup.py sdist'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        self.clear_builds()

        sys.exit()

class InstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        install.run(self)
        conf_file = Path('/etc/sre/sre.conf')
        self.execute(self.setup_click_autocompletion, args=tuple([]), msg="Setup Click Completion")
        self.rename_config_files()
        self.make_default_config(conf_file)
        self.setup_service()
        self.install_executable(conf_file)

    def rename_config_files(self):
        # Rename old config file
        path = Path('/etc/sre/autobot.conf')
        try:
            if path.exists():
                path.rename('/etc/sre/sre.conf')
        except:
            pass

    def make_default_config(self, conf_file):
        if not conf_file.exists():
            conf_file.parent.mkdir(parents=True, exist_ok=True)
            conf_file.write_text(json.dumps({
                "bots-paths": [
                ],
                "broker": {
                    "ip": "127.0.0.1",
                    "port": 1883,
                },
                "name": socket.gethostname(),
                "log_level": "info",
                "log_file": "/var/log/sre/sre.log",
                # used for webhook trigger
                "http_address": "0.0.0.0",
                "http_port": 8520,
            }, indent=4))

    def setup_service(self):
        print("Install bot as a service with 'sre install'")
        print("Pick a name for the bot service if you run several bots.")

    def setup_click_autocompletion(self):
        for console_script in setup_cfg['options']['entry_points']['console_scripts']:
            console_call = console_script.split("=")[0].strip()

            subprocess.run(["pip3", "install", "click-completion-helper", "--no-binary=click-completion-helper"])
            subprocess.run([ "click-completion-helper", "setup", console_call])

    def install_executable(self, conf_file):
        for console_script in setup_cfg['options']['entry_points']['console_scripts']:
            console_call = console_script.split("=")[0].strip()

            orig_file = Path(self.__dict__['install_scripts']) / console_call
            template = orig_file.read_text().split("\n")[1:]
            template = ["#!" + sys.executable] + template
            template = '\n'.join(template)

            bin = Path("/usr/local/bin/sre")
            bin.write_text(template)
            os.chmod(bin, 0o555)

            subprocess.run(["pip3", "install", "click-completion-helper", "--no-binary=click-completion-helper"])
            subprocess.run(['click-completion-helper', 'setup', bin.name])
            break # only one console supported

setup(
    version=about['__version__'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    # If your package is a single module, use this instead of 'packages':
    #py_modules=['srebot'],
    package_data={
        "": ["datafiles/*"],
    },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    cmdclass={
        'upload': UploadCommand,
        'install': InstallCommand,
    },
)
